import { DeferredTask } from './task.js';
import { RemoteError, TimeoutError } from './errors.js';
export async function CreateClient(url, timeout = 5000) {
    var _a;
    globalThis.WebSocket = (_a = globalThis.WebSocket) !== null && _a !== void 0 ? _a : (await import('ws')).WebSocket;
    return new class extends WebSocket {
        constructor() {
            super(url);
            this.callbacks = new Map();
            this.rpc = {};
            this.rpc = new Proxy(this.rpc, {
                get: (_instance, property, _proxy) => {
                    const method = property;
                    return (...parameters) => {
                        const reference = `${property}#${Date.now()}${Math.random()}`;
                        const message = { method, reference, parameters };
                        try {
                            const task = new DeferredTask();
                            this.callbacks.set(message.reference, task);
                            setTimeout(() => task.SetError(new TimeoutError(`Failed to perform the RPC roundtrip for '${message.reference}' within the given timeout of ${timeout}ms!`)), timeout);
                            this.send(JSON.stringify(message));
                            return task.Promise;
                        }
                        catch (error) {
                            this.callbacks.delete(message.reference);
                            throw error;
                        }
                    };
                }
            });
            this.onmessage = async (event) => {
                try {
                    const message = JSON.parse(event.data);
                    const { type, reference, result, error } = message;
                    if (type === "callback" && this.callbacks.has(reference)) {
                        const callback = this.callbacks.get(reference);
                        try {
                            if (callback && error !== undefined) {
                                return callback.SetError(new RemoteError(error));
                            }
                            if (callback && result !== undefined) {
                                return callback.SetResult(result);
                            }
                        }
                        finally {
                            this.callbacks.delete(reference);
                        }
                    }
                    throw new Error(`Failed to handle the received RPC message: '${JSON.stringify(message)}'`);
                }
                catch (error) {
                    console.warn(error);
                }
            };
        }
        get RPC() {
            return this.rpc;
        }
    }();
}
